/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.shardingsphere.integration.data.pipeline.env;

import com.google.gson.Gson;
import lombok.Getter;
import lombok.SneakyThrows;
import org.apache.shardingsphere.data.pipeline.api.config.rulealtered.yaml.YamlRuleAlteredJobConfiguration;
import org.apache.shardingsphere.data.pipeline.api.datasource.config.PipelineDataSourceConfiguration;
import org.apache.shardingsphere.data.pipeline.api.datasource.config.yaml.YamlPipelineDataSourceConfiguration;
import org.apache.shardingsphere.data.pipeline.core.datasource.PipelineDataSourceFactory;
import org.apache.shardingsphere.integration.data.pipeline.env.cases.DataSet;
import org.apache.shardingsphere.integration.data.pipeline.env.cases.Type;
import org.apache.shardingsphere.integration.data.pipeline.env.config.SourceConfiguration;
import org.apache.shardingsphere.integration.data.pipeline.env.config.TargetConfiguration;
import org.apache.shardingsphere.sharding.yaml.config.rule.YamlTableRuleConfiguration;

import javax.sql.DataSource;
import javax.xml.bind.JAXBContext;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Map;

/**
 * Integration test environment context.
 */
@Getter
public final class ITEnvironmentContext {
    
    public static final ITEnvironmentContext INSTANCE = new ITEnvironmentContext();
    
    private static final String TYPE_TEST_XML = "/cases/mysql/types.xml";
    
    private final DataSet testCases;
    
    private final DataSource sourceDataSource;
    
    private final DataSource targetDataSource;
    
    private final String scalingConfiguration;
    
    public ITEnvironmentContext() {
        testCases = loadTestCases();
        Map<String, YamlTableRuleConfiguration> sourceTableRules = createSourceTableRules();
        scalingConfiguration = createScalingConfiguration(sourceTableRules);
        sourceDataSource = SourceConfiguration.createHostDataSource(sourceTableRules);
        targetDataSource = PipelineDataSourceFactory.newInstance(TargetConfiguration.getHostConfiguration());
    }
    
    @SneakyThrows
    private DataSet loadTestCases() {
        try (FileReader reader = new FileReader(ITEnvironmentContext.class.getResource(TYPE_TEST_XML).getPath())) {
            return (DataSet) JAXBContext.newInstance(DataSet.class).createUnmarshaller().unmarshal(reader);
        }
    }
    
    private Map<String, YamlTableRuleConfiguration> createSourceTableRules() {
        Map<String, YamlTableRuleConfiguration> result = new HashMap<>(testCases.getTypes().size(), 1);
        for (Type type : testCases.getTypes()) {
            YamlTableRuleConfiguration tableRule = new YamlTableRuleConfiguration();
            tableRule.setLogicTable(type.getTableName());
            tableRule.setActualDataNodes("ds_src." + type.getTableName());
            result.put(type.getTableName(), tableRule);
        }
        return result;
    }
    
    private static String createScalingConfiguration(final Map<String, YamlTableRuleConfiguration> tableRules) {
        YamlRuleAlteredJobConfiguration jobConfig = new YamlRuleAlteredJobConfiguration();
        jobConfig.setSource(createYamlPipelineDataSourceConfiguration(SourceConfiguration.getDockerConfiguration(tableRules)));
        jobConfig.setTarget(createYamlPipelineDataSourceConfiguration(TargetConfiguration.getDockerConfiguration()));
        return new Gson().toJson(jobConfig);
    }
    
    private static YamlPipelineDataSourceConfiguration createYamlPipelineDataSourceConfiguration(final PipelineDataSourceConfiguration targetConfig) {
        YamlPipelineDataSourceConfiguration result = new YamlPipelineDataSourceConfiguration();
        result.setType(targetConfig.getType());
        result.setParameter(targetConfig.getParameter());
        return result;
    }
}
